

from artemis.experiments.experiment_record import *

import logging

logging.getLogger(__name__).warning('The module artemis.fileman.experiment_record is deprecated and will eventually be removed.  Import artemis.experiments.experiment_record instead.')
